local _,stash = ...
local C = stash.constants

--[[
    This file is for tracking itemIDs as a player acquires/loses them and where they are.

    Expected behaviors:
    - For various reasons, this addon only tracks itemIDs and not any customization done on
      items. For instance, 6/6 Korthian armor will be their base armor. Plate Boots of the Eagle
      will be just Plate Boots without the suffix. Enchants, gems, etc. will not be attached to
      any item.
    - On Classic realms, sending mail to another character still takes an hour. If you check
      mail on the other character before it's arrived, it will remove the incoming item
      from the other character's mail stash (until you check mail when it's arrived). This addon
      does not attempt to track time it takes for mail to arrive.

]]

StashSaved = {} -- everything is saved in StashSaved savedvar, indexed by realm and then character name

local dirtyBags = {[0]=true,[1]=true,[2]=true,[3]=true,[4]=true} -- bag slots (0-4) that change in BAG_UPDATE (login as all dirty)
local dirtyBank = {[-1]=true,[5]=true,[6]=true,[7]=true,[8]=true,[9]=true,[10]=true,[11]=true} -- bank bags (5-11) that change in BAG_UPDATE

local isBankOpen -- true while bank is open

local callbacks = {} -- indexed by "item" or "money", functions to call when items or money has changed

local pendingSendMail = {} -- ordered list with first row as {realm,recipient} and rest of rows as {itemID,count}

-- self.tracking is an independent frame solely for item tracking events and methods
stash.tracking = CreateFrame("Frame") 
stash.tracking:SetScript("OnEvent",function(self,event,...)
    if self[event] then
        self[event](self,...)
    end
    if callbacks.item then
        callbacks.item()
    end
end)

-- housekeeping done once on login
function stash.tracking:Setup()
    self:RegisterEvent("BAG_UPDATE")
    self:RegisterEvent("BAG_UPDATE_DELAYED")
    self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED")
    self:RegisterEvent("ITEM_UNLOCKED")
    self:RegisterEvent("PLAYER_MONEY")
    self:RegisterEvent("BANKFRAME_OPENED")
    self:RegisterEvent("BANKFRAME_CLOSED")
    self:RegisterEvent("PLAYERBANKSLOTS_CHANGED")

    self:RegisterEvent("MAIL_CLOSED")
    self:RegisterEvent("MAIL_SEND_SUCCESS")
    self:RegisterEvent("MAIL_INBOX_UPDATE")

    if not C.IS_CLASSIC then
        self:RegisterEvent("PLAYERREAGENTBANKSLOTS_CHANGED")
        self:RegisterEvent("CURRENCY_DISPLAY_UPDATE")
    end

    -- on login, take an inventory of accessible items and money on character
    self:BAG_UPDATE_DELAYED()
    for slot=1,23 do -- and update worn stuff
        self:PLAYER_EQUIPMENT_CHANGED(slot)
    end
    if not C.IS_CLASSIC then
        self:UpdateContainer(stash.current.reagents,-3) -- reagents are always available it seems?
        self:UpdateAllCurrency() -- updates player's currency
        self:VOID_TRANSFER_DONE()
    end
    self:PLAYER_MONEY()

    hooksecurefunc("SendMail",stash.tracking.SendMailHook) -- watch for mail being sent
end

-- when either "item" or "money" changes in tracking, call the given function. multiple
-- modules can call this and it will run their functions for the tracking type in the order
-- they were "registered"
function stash.tracking:SetCallback(callbackType,func)
    if type(func)~="function" or (callbackType~="item" and callbackType~="money") then
        return
    end
    if callbacks[callbackType] then
        local old = callbacks[callbackType]
        callbacks[callbackType] = function()
            old()
            func()
        end
    else
        callbacks[callbackType] = func
    end
end

--[[ events ]]

function stash.tracking:BAG_UPDATE(bag)
    if bag>=0 and bag<=4 then -- for worn bags that are changing
        dirtyBags[bag] = true
    elseif bag~=-2 then -- otherwise it's a bank bag that needs to wait until bank open
        dirtyBank[bag] = true
    end
end

-- fired after a set of BAG_UPDATEs; go through each bag that had an update and update their contents
function stash.tracking:BAG_UPDATE_DELAYED()
    for bag in pairs(dirtyBags) do
        stash.tracking:UpdateContainer(stash.current.bags,bag)
    end
    wipe(dirtyBags)
    if isBankOpen then
        for bag in pairs(dirtyBank) do
            stash.tracking:UpdateContainer(stash.current.bank,bag)
        end
        wipe(dirtyBank)
    end
end

-- nothing stacked can be worn (ammo in classic exists in bags)
function stash.tracking:PLAYER_EQUIPMENT_CHANGED(slot)
    stash.current.worn[slot] = GetInventoryItemID("player",slot)
end

-- worn bags are inventory slots 20-23
function stash.tracking:ITEM_UNLOCKED(slot)
    if slot and slot>=20 and slot<=23 then
        stash.current.worn[slot] = GetInventoryItemID("player",slot)
    end
end

-- when player's money changes; this value is an integer of copper amount
function stash.tracking:PLAYER_MONEY()
    stash.current.money = GetMoney()
    if callbacks.money then
        callbacks.money()
    end
end

-- there is no equivalent of BAG_UPDATE_DELAYED for reagent bank tab, just a per-slot change
function stash.tracking:PLAYERREAGENTBANKSLOTS_CHANGED(slot)
    self:UpdateContainerSlot(stash.current.reagents,-3,slot)
end

-- this is for the main bank area (bag slot -1) and in retail can fire if quest/tradeskill items used; but if the bank
-- is not open then we can't scan changes yet
function stash.tracking:PLAYERBANKSLOTS_CHANGED(slot)
    if isBankOpen then
        self:UpdateContainerSlot(stash.current.bank,-1,slot)
        dirtyBank[-1] = true -- kinda redundant, since whole bank scanned when opened; but being safe
    end
end

-- when currency changes, can directly update the changing currency
function stash.tracking:CURRENCY_DISPLAY_UPDATE(currencyID,quantity)
    stash.current.currency[format("currency:%d",currencyID)] = quantity
    if callbacks.item then
        callbacks.item()
    end
end

function stash.tracking:BANKFRAME_OPENED()
    isBankOpen = true
    -- it's possible that reagents or quest items in bank were consumed, rescan whole bank when it's opened
    for bag in pairs(stash.current.bank) do
        dirtyBank[bag] = true
    end
    -- after a fraction of a second, pick up bank contents
    C_Timer.After(C.BANK_OPEN_DELAY,function()
        stash.tracking:BAG_UPDATE_DELAYED()
        if callbacks.item then
            callbacks.item()
        end
    end)
end

function stash.tracking:BANKFRAME_CLOSED()
    isBankOpen = false
end

--[[ helper methods for events ]]

-- for bags, bank or reagents, updates all contents of the given bag in the given savedvar table
-- we need to get use GetContainerItemInfo to get a count, but it's not valid if an itemID isn't data cached
-- container is one of: stash.current.bags, stash.current.bank or stash.current.reagents
-- bag is 0-4 for bags, 5-11 for bank, -3 for reagents
function stash.tracking:UpdateContainer(container,bag)
    if container==stash.current.bank and not isBankOpen then
        return -- we were about to update bank bags while bank isn't open; leave
    end
    local needsDataCache = false -- becomes true if anything isn't data cached
    if container[bag] then
        wipe(container[bag])
    end
    for slot=1,GetContainerNumSlots(bag) do
        if not self:UpdateContainerSlot(container,bag,slot) then
            needsDataCache = true
        end
    end
    if needsDataCache then -- if something wasn't data cached, come back shortly and try again
        C_Timer.After(C.DATA_CACHE_DELAY,function() -- using a closure here so the same container,bag is retried
            stash.tracking:UpdateContainer(container,bag)
            if callbacks.item then
                callbacks.item()
            end
        end)
    end
end

-- updates the bag,slot of the given container and returns true if it was successfully cached; false if a data cache retry needs to happen
-- we need to use GetContainerItemInfo to get a count, but it's not valid if an itemID isn't data cached
function stash.tracking:UpdateContainerSlot(container,bag,slot)
    local itemID = GetContainerItemID(bag,slot) -- this value is reliable even if not data cached
    if itemID then -- only filling slots that have an item
        if C_Item.IsItemDataCachedByID(itemID) then -- if item data cached, then its count is reliable
            local _,count,_,_,_,_,link = GetContainerItemInfo(bag,slot)
            if itemID==C.PET_CAGE_ITEMID then -- this is a caged pet which fortunately never stacks, save it as "battlepet:speciesID"
                container[bag][slot] = (link or ""):match("(battlepet:%d+)") or itemID
            elseif count>1 then
                container[bag][slot] = format("%d:%d",itemID,count)
            else
                container[bag][slot] = itemID
            end
            return true
        else -- itemID isn't data cached, which we need to get a count of items in the bag,slot; request data
            C_Item.RequestLoadItemDataByID(itemID)
            return false
        end
    elseif container[bag] then -- no itemID, empty it out
        container[bag][slot] = nil
        return true
    end
end

-- updates all currency for current character
local collapsedCurrencies = {} -- lookup table to backup collapsed currency headers
function stash.tracking:UpdateAllCurrency()
    -- backup collapsed headers
    for i=1,C_CurrencyInfo.GetCurrencyListSize() do
        local info = C_CurrencyInfo.GetCurrencyListInfo(i)
        if info.isHeader and not info.isHeaderExpanded then
            collapsedCurrencies[info.name] = true
        end
    end
    -- expand all headers
    for i=C_CurrencyInfo.GetCurrencyListSize(),1,-1 do
        local info = C_CurrencyInfo.GetCurrencyListInfo(i)
        if collapsedCurrencies[info.name] then
          C_CurrencyInfo.ExpandCurrencyList(i,true)
        end
    end
    -- gather currency
    wipe(stash.current.currency)
    for i=1,C_CurrencyInfo.GetCurrencyListSize() do
        local info = C_CurrencyInfo.GetCurrencyListInfo(i)
        if not info.isHeader and info.quantity and info.quantity>0 then
            local link = (C_CurrencyInfo.GetCurrencyListLink(i) or ""):match("(currency:%d+)")
            if link then
                stash.current.currency[link] = info.quantity
            end
        end
    end
    -- restore headers
    for i=C_CurrencyInfo.GetCurrencyListSize(),1,-1 do
        local info = C_CurrencyInfo.GetCurrencyListInfo(i)
        if collapsedCurrencies[info.name] then
            C_CurrencyInfo.ExpandCurrencyList(i,false)
        end
    end
        
    wipe(collapsedCurrencies)
end

-- mail

-- when mail is sent, if any items are involved then save them in pendingSendMail
-- when/if mail is successfully sent, these items will be added to recipient's stash if
-- they have a stash.saved node set up.
-- note the dot notation! this is a hook of a Blizzard API without a self reference
function stash.tracking.SendMailHook(recipient)
    wipe(pendingSendMail)
    -- add any items being sent to pendingSendMail with its count
    for i=1,ATTACHMENTS_MAX_SEND do
        if HasSendMailItem(i) then
            local _,itemID,_,count = GetSendMailItem(i)
            if itemID==C.PET_CAGE_ITEMID then -- item is a caged pet
                local link = GetSendMailItemLink(i)
                itemID = link:match("(battlepet:%d+)")
                if itemID then
                    tinsert(pendingSendMail,itemID)
                end
            elseif count>1 then
                tinsert(pendingSendMail,format("%d:%d",itemID,count))
            elseif itemID  then
                tinsert(pendingSendMail,itemID)
            end
        end
    end
    -- if an item is being sent, insert the realm,player the item(s) are being sent to
    if #pendingSendMail>0 then
        local recipientRealm = GetRealmName()
        if recipient:match("-") then -- send a BoA item to another realm
            recipient,recipientRealm = recipient:match("(.+)-(.+)")
        end
        tinsert(pendingSendMail,1,{recipientRealm,recipient})
    end
end

-- if mail was successfully sent, check if any pending items were being sent
function stash.tracking:MAIL_SEND_SUCCESS()
    if #pendingSendMail>1 then
        local realm,player = pendingSendMail[1][1],pendingSendMail[1][2]
        local otherStash = stash.saved[realm] and stash.saved[realm][player]
        if otherStash then
            for i=2,#pendingSendMail do
                tinsert(otherStash.mail,pendingSendMail[i])
            end
        end
    end
    wipe(pendingSendMail)
end

-- when mail closes, any items in SendMail are released
function stash.tracking:MAIL_CLOSED()
    wipe(pendingSendMail)
end

-- when opening mail, update the current player's mail
function stash.tracking:MAIL_INBOX_UPDATE()
    wipe(stash.current.mail) -- we're checking mail now, can drop stuff that was noted as sent
    for i=1,GetInboxNumItems() do
        if select(8,GetInboxHeaderInfo(i)) then -- if this inbox item has items
            for j=1,ATTACHMENTS_MAX_RECEIVE do
                if HasInboxItem(i,j) then
                    local _,itemID,_,count = GetInboxItem(i,j)
                    if itemID==C.PET_CAGE_ITEMID then -- item is a caged pet
                        -- GetInboxItemLink doesn't use a battlepet link :( so picking up speciesID from tooltip
                        StashGameTooltip:SetOwner(stash.frame,"ANCHOR_NONE")
                        local _,speciesID = StashGameTooltip:SetInboxItem(i,j)
                        if speciesID and speciesID>0 then
                            tinsert(stash.current.mail,format("battlepet:%d",speciesID))
                        end
                        StashGameTooltip:Hide()
                    elseif count>1 then
                        tinsert(stash.current.mail,format("%d:%d",itemID,count))
                    elseif itemID  then
                        tinsert(stash.current.mail,itemID)
                    end
                end
            end
        end
    end
end

function stash.tracking:VOID_TRANSFER_DONE()
    if not CanUseVoidStorage() then
        return -- player doesn't have voide storage unlocked, leave
    end
    for tab=1,C.VOID_STORAGE_MAX_TABS do
        for slot=1,C.VOID_STORAGE_MAX_SLOTS do
            local itemID = GetVoidItemInfo(tab,slot)
            stash.current.void[tab][slot] = itemID
        end
    end
end
